import os
from pathlib import Path

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent

SECRET_KEY = "django-insecure-%am#r-wl5tx8vp6+l7o+6)3+1iipnex@otx_%dtov7gbnud5t0"

DEBUG = True

# Application definition

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    # First party
    "itoutils.django",
    # First party's tests
    "testproject.testapp",
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "testproject.testapp.nexus.middleware.AutoLoginMiddleware",
]

ROOT_URLCONF = "testproject.config.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

# Database

DATABASES = {
    "default": {
        "ATOMIC_REQUESTS": True,
        "ENGINE": "django.db.backends.postgresql",
        "NAME": os.getenv("PGDATABASE", "itoutils"),
        "HOST": os.getenv("PGHOST", "127.0.0.1"),
        "PORT": os.getenv("PGPORT", "5432"),
        "USER": os.getenv("PGUSER", "postgres"),
        "PASSWORD": os.getenv("PGPASSWORD", "password"),
    }
}

# Static files (CSS, JavaScript, Images)

STATIC_URL = "static/"

# Default primary key field type

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"

# Logging

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "json": {"()": "itoutils.django.logging.DataDogJSONFormatter"},
    },
    "handlers": {
        "console": {"class": "logging.StreamHandler", "formatter": "json"},
        "null": {"class": "logging.NullHandler"},
    },
    "loggers": {
        "": {"handlers": ["console"], "level": "INFO"},
    },
}


NEXUS_AUTO_LOGIN_KEY = {"k": "aTR4ZnR1WlpYYmphbFdtaXVlVjB3alljNjhrWXpfYSE", "kty": "oct"}
NEXUS_ALLOWED_REDIRECT_HOSTS = ["plateforme.inclusion.beta.gouv.fr", "plateforme.inclusion.gouv.fr"]



import logging

from django.contrib.auth import get_user_model, logout
from django.http import HttpResponseRedirect
from django.utils.deprecation import MiddlewareMixin

from itoutils.django.nexus.token import decode_token

logger = logging.getLogger(__name__)


class BaseAutoLoginMiddleware(MiddlewareMixin):
    def get_queryset(self):
        return get_user_model().objects.all()

    def get_proconnect_authorize_url(self, user, next_url):
        raise NotImplementedError

    def get_no_user_url(self, email, next_url):
        raise NotImplementedError

    def process_request(self, request):
        if "auto_login" not in request.GET:
            return

        query_params = request.GET.copy()
        auto_login = query_params.pop("auto_login")
        email = None

        new_url = f"{request.path}?{query_params.urlencode()}" if query_params else request.path

        if len(auto_login) != 1:
            logger.info("Nexus auto login: Multiple tokens found -> ignored")
            return HttpResponseRedirect(new_url)
        else:
            [token] = auto_login

        try:
            decoded_data = decode_token(token)
            email = decoded_data.get("email")
        except ValueError:
            logger.info("Invalid auto login token")

        if email is None:
            logger.info("Nexus auto login: Missing email in token -> ignored")
            return HttpResponseRedirect(new_url)

        if request.user.is_authenticated:
            if request.user.email == email:
                logger.info("Nexus auto login: user is already logged in")
                return HttpResponseRedirect(new_url)
            else:
                logger.info("Nexus auto login: wrong user is logged in -> logging them out")
                # We should probably also logout the user from ProConnect but it requires to redirect to ProConnect
                # and the flow becomes a lot more complicated
                logout(request)

        try:
            user = self.get_queryset().get(email=email)
            logger.info("Nexus auto login: %s was found and forwarded to ProConnect", user)
            return HttpResponseRedirect(self.get_proconnect_authorize_url(user, new_url))
        except get_user_model().DoesNotExist:
            logger.info("Nexus auto login: no user found for jwt=%s", token)
            return HttpResponseRedirect(self.get_no_user_url(email, new_url))



from django.core.management import BaseCommand

from itoutils.django.commands import LoggedCommandMixin, dry_runnable
from testproject.testapp.models import Item


class Command(LoggedCommandMixin, BaseCommand):
    ATOMIC_HANDLE = True

    def add_arguments(self, parser):
        parser.add_argument("pk", type=int)
        parser.add_argument("--delete", dest="delete", action="store_true")
        parser.add_argument("--wet-run", dest="wet_run", action="store_true")

    @dry_runnable
    def handle(self, pk, delete=False, **options):
        item = Item.objects.get(pk=pk)
        if delete:
            self.logger.info("Deleting Item pk=%d", pk)
            item.delete()
        else:
            print(item)



import json
import logging
import time

from django.conf import settings
from jwcrypto import jwk, jwt

logger = logging.getLogger(__name__)

EXPIRY_DELAY = 60  # seconds


# Use a function to easily override the setting in tests
def _get_key():
    return jwk.JWK(**settings.NEXUS_AUTO_LOGIN_KEY) if settings.NEXUS_AUTO_LOGIN_KEY else None


def generate_token(user):
    token = jwt.JWT(
        header={"alg": "A256KW", "enc": "A256CBC-HS512"},
        claims={"email": user.email, "exp": round(time.time()) + EXPIRY_DELAY},
    )
    token.make_encrypted_token(_get_key())
    return token.serialize()


def decode_token(token):
    try:
        claims = json.loads(jwt.JWT(key=_get_key(), jwt=token, expected_type="JWE").claims)
        claims.pop("exp", None)
        return claims
    except Exception as err:
        logger.exception("Could not decrypt jwt")
        raise ValueError from err



from urllib.parse import parse_qsl, urlencode, urlsplit


def add_url_params(url: str, params: dict[str, str]) -> str:
    """Add GET params to provided URL being aware of existing.

    :param url: string of target URL
    :param params: dict containing requested params to be added
    :return: string with updated URL

    >> url = 'http://localhost:8000/login/activate_employer_account?next_url=%2Finvitations
    >> new_params = {'test': 'value' }
    >> add_url_params(url, new_params)
    'http://localhost:8000/login/activate_employer_account?next_url=%2Finvitations&test=value
    """

    # Remove params with None values
    params = {key: value for key, value in params.items() if value is not None}

    url_parts = urlsplit(url)
    query = dict(parse_qsl(url_parts.query))
    query.update(params)

    new_url = url_parts._replace(query=urlencode(query)).geturl()

    return new_url



import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Item",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "parent",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="children",
                        to="testapp.item",
                    ),
                ),
            ],
        ),
    ]



import pytest
from django.contrib.auth.models import User
from pytest_django.asserts import assertRedirects

from itoutils.django.nexus.token import generate_token
from itoutils.urls import add_url_params
from tests.django.factories import UserFactory


@pytest.fixture(params=[{}, {"filter": "76", "username": "123"}], ids=("no_params", "with_params"))
def params(request):
    return request.param


def test_middleware_for_authenticated_user(db, client, params, caplog):
    expected_url = add_url_params("/path", params)

    user = UserFactory()
    client.force_login(user)

    response = client.get(add_url_params("/path", params | {"auto_login": generate_token(user)}))
    assertRedirects(response, expected_url, fetch_redirect_response=False)
    assert caplog.messages == ["Nexus auto login: user is already logged in"]


def test_middleware_for_wrong_authenticated_user(db, client, params, caplog):
    expected_url = add_url_params("/path", params)

    user = UserFactory()
    # Another user is logged in
    client.force_login(User.objects.create(email="moi@mailinator.com", pk=2))

    response = client.get(add_url_params("/path", params | {"auto_login": generate_token(user)}))
    assertRedirects(
        response,
        add_url_params("/proconnect/authorize", {"email": user.email, "next_url": expected_url}),
        fetch_redirect_response=False,
    )
    assert caplog.messages == [
        "Nexus auto login: wrong user is logged in -> logging them out",
        f"Nexus auto login: {user} was found and forwarded to ProConnect",
    ]


def test_middleware_multiple_tokens(db, client, caplog):
    user = UserFactory()
    token = generate_token(user)
    response = client.get(f"/path?auto_login={token}&auto_login={token}")
    assertRedirects(response, "/path", fetch_redirect_response=False)
    assert caplog.messages == [
        "Nexus auto login: Multiple tokens found -> ignored",
    ]


def test_middleware_invalid_token(db, client, caplog):
    params = {"auto_login": "bad jwt"}
    response = client.get(add_url_params("/path", params))
    assertRedirects(response, "/path", fetch_redirect_response=False)
    assert caplog.messages == [
        "Could not decrypt jwt",
        "Invalid auto login token",
        "Nexus auto login: Missing email in token -> ignored",
    ]


def test_middleware_with_no_existing_user(db, client, params, caplog):
    expected_url = add_url_params("/path", params)

    user = UserFactory.build()
    jwt = generate_token(user)
    response = client.get(add_url_params("/path", params | {"auto_login": jwt}))
    assertRedirects(
        response,
        add_url_params("/register", {"email": user.email, "next_url": expected_url}),
        fetch_redirect_response=False,
    )
    assert caplog.messages == [f"Nexus auto login: no user found for jwt={jwt}"]


def test_middleware_for_unlogged_user(db, client, params, caplog):
    expected_url = add_url_params("/path", params)

    user = UserFactory()

    response = client.get(add_url_params("/path", params | {"auto_login": generate_token(user)}))
    assertRedirects(
        response,
        add_url_params("/proconnect/authorize", {"email": user.email, "next_url": expected_url}),
        fetch_redirect_response=False,
    )
    assert caplog.messages == [f"Nexus auto login: {user} was found and forwarded to ProConnect"]



import json
import logging
import uuid

import pytest
from django.core import management
from django.db import IntegrityError
from pytest_django.asserts import assertQuerySetEqual

from itoutils.django.commands import dry_runnable
from testproject.testapp.models import Item


def test_current_command_info_are_logged(capture_stream_handler_log):
    with capture_stream_handler_log(logging.getLogger()) as captured:
        management.call_command(
            "echo",
            "Hello World",
        )
    # Check that the command's info are properly logged
    log = json.loads(captured.getvalue().splitlines()[0])
    assert log["command.name"] == "testproject.testapp.management.commands.echo"
    assert uuid.UUID(log["command.run_uid"])


@pytest.mark.parametrize(
    "command,expected",
    [
        # Command with a wet_run argument
        ("testproject.testapp.management.commands.item", False),
        ("testproject.testapp.management.commands.item", True),
        # Command without a wet_run argument
        ("testproject.testapp.management.commands.echo", None),
    ],
)
@pytest.mark.django_db
def test_wet_run_info_is_logged(capture_stream_handler_log, command, expected, item):
    args = ["--wet-run"] if expected else []
    with capture_stream_handler_log(logging.getLogger()) as captured:
        management.call_command(
            # This could have been any other command inheriting from LoggedCommandMixin
            command.rsplit(".", 1)[-1],
            item.pk,
            *args,
        )
    # Check that wet run information is properly logged
    log = json.loads(captured.getvalue().splitlines()[0])
    # Check extra
    assert log["command.wet_run"] is expected
    # Check log
    if expected is False:
        assert "Command launched with wet_run=False" in captured.getvalue()
        assert "Setting transaction to be rollback as wet_run=False" in captured.getvalue()


@pytest.mark.django_db
def test_dry_runnable(item):
    assertQuerySetEqual(Item.objects.all(), [item])
    management.call_command(
        "item",
        item.pk,
        "--delete",
    )
    assertQuerySetEqual(Item.objects.all(), [item])
    management.call_command(
        "item",
        item.pk,
        "--delete",
        "--wet-run",
    )
    assertQuerySetEqual(Item.objects.all(), [])


@pytest.mark.django_db
def test_dry_runnable_check_constraints(item):
    @dry_runnable
    def command(**kwargs):
        item.parent_id = item.pk + 1
        item.save(update_fields={"parent_id"})

    with pytest.raises(
        IntegrityError,
        match='violates foreign key constraint "testapp_item_parent_id_.*_fk_testapp_item_id"',
    ):
        command(wet_run=False)



from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.http import Http404, HttpResponseRedirect
from django.utils.http import url_has_allowed_host_and_scheme

from itoutils.django.nexus.token import generate_token
from itoutils.urls import add_url_params


@login_required
def auto_login(request):
    next_url = request.GET.get("next_url")

    if next_url is None or settings.NEXUS_AUTO_LOGIN_KEY is None:
        raise Http404

    if url_has_allowed_host_and_scheme(next_url, settings.NEXUS_ALLOWED_REDIRECT_HOSTS, require_https=True):
        return HttpResponseRedirect(add_url_params(next_url, {"auto_login": generate_token(request.user)}))

    raise Http404



from django.db import models


class Item(models.Model):
    parent = models.ForeignKey("self", on_delete=models.CASCADE, related_name="children", null=True)

    class Meta:
        app_label = "testapp"



from itoutils.django.nexus.middleware import BaseAutoLoginMiddleware
from itoutils.urls import add_url_params


class AutoLoginMiddleware(BaseAutoLoginMiddleware):
    def get_proconnect_authorize_url(self, user, next_url):
        return add_url_params("/proconnect/authorize", {"email": user.email, "next_url": next_url})

    def get_no_user_url(self, email, next_url):
        return add_url_params("/register", {"email": email, "next_url": next_url})



from django.conf import settings
from django.test import override_settings
from django.urls import reverse
from pytest_django.asserts import assertRedirects

from itoutils.urls import add_url_params
from tests.django.factories import UserFactory


class TestAutoLogin:
    def test_login_required(self, db, client):
        next_url = f"https://{settings.NEXUS_ALLOWED_REDIRECT_HOSTS[0]}"
        url = reverse("nexus:auto_login", query={"next_url": next_url})
        response = client.get(url)
        assertRedirects(response, add_url_params(settings.LOGIN_URL, {"next": url}), fetch_redirect_response=False)

    def test_nominal_case(self, db, client, mock_nexus_token):
        client.force_login(UserFactory())
        for host in settings.NEXUS_ALLOWED_REDIRECT_HOSTS:
            next_url = f"https://{host}"
            url = reverse("nexus:auto_login", query={"next_url": next_url})
            response = client.get(url)
            assertRedirects(response, add_url_params(next_url, {"auto_login": "JWT"}), fetch_redirect_response=False)

    def test_missing_next_url(self, db, client):
        client.force_login(UserFactory())
        # Without next_url
        url = reverse("nexus:auto_login")
        response = client.get(url)
        assert response.status_code == 404

    def test_bad_host(self, db, client):
        client.force_login(UserFactory())
        next_url = "https://empl0is.fr"
        url = reverse("nexus:auto_login", query={"next_url": next_url})
        response = client.get(url)
        assert response.status_code == 404

    @override_settings(NEXUS_AUTO_LOGIN_KEY=None)
    def test_no_settings(self, db, client):
        client.force_login(UserFactory())
        next_url = f"https://{settings.NEXUS_ALLOWED_REDIRECT_HOSTS[0]}"
        url = reverse("nexus:auto_login", query={"next_url": next_url})
        response = client.get(url)
        assert response.status_code == 404



import collections
import contextlib
import functools
import logging
import time
import uuid

from asgiref.local import Local
from django.db import connection, transaction

local = Local()

CommandInfo = collections.namedtuple("CommandInfo", ["run_uid", "name", "wet_run"])


def get_current_command_info():
    return getattr(local, "command_info", None)


def _log_command_result(command, duration_in_ns, result):
    command.logger.info(
        f"Management command %s {result} in %0.2f seconds",
        command.__module__,
        duration_in_ns / 1_000_000_000,
        extra={
            "command": command.__module__,
            # Datadog expects duration in ns
            "duration": duration_in_ns,
        },
    )


@contextlib.contextmanager
def _command_duration_logger(command):
    before = time.perf_counter_ns()
    try:
        yield
    except Exception:
        _log_command_result(command, time.perf_counter_ns() - before, "failed")
        raise
    _log_command_result(command, time.perf_counter_ns() - before, "succeeded")


@contextlib.contextmanager
def _command_info_manager(command, *, wet_run=None):
    command_info_before = get_current_command_info()
    run_uid = command_info_before.run_uid if command_info_before else str(uuid.uuid4())
    local.command_info = CommandInfo(run_uid, command.__class__.__module__, wet_run)
    try:
        yield local.command_info
    finally:
        local.command_info = command_info_before


class LoggedCommandMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(self.__class__.__module__)

    def execute(self, *args, **kwargs):
        with (
            _command_info_manager(self, wet_run=kwargs.get("wet_run")),
            _command_duration_logger(self),
        ):
            try:
                return super().execute(*args, **kwargs)
            except Exception:
                self.logger.exception(
                    "Error when executing %s",
                    self.__module__,
                    extra={
                        "command": self.__module__,
                    },
                )
                raise


class AtomicHandleMixin:
    ATOMIC_HANDLE = False

    def execute(self, *args, **kwargs):
        with transaction.atomic() if self.ATOMIC_HANDLE else contextlib.nullcontext():
            return super().execute(*args, **kwargs)


def dry_runnable(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        wet_run = kwargs.get("wet_run")
        if wet_run is None:
            raise RuntimeError('No "wet_run" argument was given')

        logger = (
            args[0].logger
            if args and args[0] and isinstance(args[0], LoggedCommandMixin)
            else logging.getLogger(func.__module__)
        )
        with transaction.atomic():
            if not wet_run:
                logger.info("Command launched with wet_run=%s", wet_run)
            func(*args, **kwargs)
            if not wet_run:
                with connection.cursor() as cursor:
                    cursor.execute("SET CONSTRAINTS ALL IMMEDIATE;")
                transaction.set_rollback(True)
                logger.info("Setting transaction to be rollback as wet_run=%s", wet_run)

    return wrapper



from django.core.management import BaseCommand

from itoutils.django.commands import LoggedCommandMixin


class Command(LoggedCommandMixin, BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument("value", type=str)

    def handle(self, value, **options):
        print(value)



import datetime

import pytest
from django.utils import timezone
from jwcrypto import jwt

from itoutils.django.nexus.token import EXPIRY_DELAY, decode_token, generate_token
from tests.django.factories import UserFactory


def test_generate_and_decode_token(db, time_machine):
    now = timezone.now()
    time_machine.move_to(now)
    user = UserFactory()
    token = generate_token(user)

    # generated token requires a key to decode
    with pytest.raises(KeyError):
        jwt.JWT(jwt=token).claims  # noqa: B018

    # It contains the user email
    assert decode_token(token) == {"email": user.email}

    # Wait for the JWT to expire, and then extra time for the leeway.
    leeway = 60
    time_machine.move_to(now + datetime.timedelta(seconds=EXPIRY_DELAY + leeway + 1))
    with pytest.raises(ValueError):
        decode_token(token)



from itoutils.urls import add_url_params


def test_add_url_params():
    base_url = "http://localhost/test?next=/path%3Fparam1%3D100%26param2%3Dabc"

    url_test = add_url_params(base_url, {"test": "value"})
    assert url_test == "http://localhost/test?next=%2Fpath%3Fparam1%3D100%26param2%3Dabc&test=value"

    url_test = add_url_params(base_url, {"mypath": "%2Fvalue%2Fpath"})

    assert url_test == ("http://localhost/test?next=%2Fpath%3Fparam1%3D100%26param2%3Dabc&mypath=%252Fvalue%252Fpath")

    url_test = add_url_params(base_url, {"mypath": None})

    assert url_test == "http://localhost/test?next=%2Fpath%3Fparam1%3D100%26param2%3Dabc"

    url_test = add_url_params(base_url, {"mypath": ""})

    assert url_test == "http://localhost/test?next=%2Fpath%3Fparam1%3D100%26param2%3Dabc&mypath="



import factory
from django.contrib.auth.models import User


class UserFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = User

    username = factory.Sequence("user_name{}".format)
    first_name = factory.Faker("first_name")
    last_name = factory.Faker("last_name")
    email = factory.Sequence("email{}@domain.com".format)



import contextlib
import io
import logging
from unittest import mock

import pytest


@pytest.fixture
def capture_stream_handler_log(request):
    # Workaround capsys/capfd not working, see https://github.com/pytest-dev/pytest/issues/5997
    @contextlib.contextmanager
    def capture_stream_handler(logger):
        with contextlib.ExitStack() as stack:
            captured = io.StringIO()
            for handler in logger.handlers:
                if isinstance(handler, logging.StreamHandler):
                    stack.enter_context(mock.patch.object(handler, "stream", captured))
            yield captured

    return capture_stream_handler


@pytest.fixture
def mock_nexus_token():
    with mock.patch("itoutils.django.nexus.views.generate_token", return_value="JWT") as mocked:
        yield mocked



import logging

from django_datadog_logger.formatters import datadog

from itoutils.django.commands import get_current_command_info

logger = logging.getLogger(__name__)


class DataDogJSONFormatter(datadog.DataDogJSONFormatter):
    # We don't want those information in our logs
    LOG_KEYS_TO_REMOVE = ["usr.name", "usr.email", "usr.session_key"]

    def json_record(self, message, extra, record):
        log_entry_dict = super().json_record(message, extra, record)
        for log_key in self.LOG_KEYS_TO_REMOVE:
            if log_key in log_entry_dict:
                del log_entry_dict[log_key]
        if (command_info := get_current_command_info()) is not None:
            log_entry_dict["command.run_uid"] = command_info.run_uid
            log_entry_dict["command.name"] = command_info.name
            log_entry_dict["command.wet_run"] = command_info.wet_run
        return log_entry_dict
